from brownie import Contract, chain
from cachetools.func import ttl_cache

from ypricemagic.utils.cache import memory
from ypricemagic.utils.multicall2 import fetch_multicall
from ypricemagic.utils.raw_calls import _decimals


@ttl_cache(ttl=600)
def get_markets():
    if chain.id == 1: # eth mainnet
        comptroller = Contract("0x3d9819210A31b4961b30EF54bE2aeD79B9c9Cd3B")
        creamtroller = Contract('0x3d5BC3c8d13dcB8bF317092d84783c2697AE9258')
        ironbankroller = Contract("0xAB1c342C7bf5Ec5F02ADEA1c2270670bCa144CbB")
        
        results = fetch_multicall(
            [comptroller, 'getAllMarkets'],
            [creamtroller, 'getAllMarkets'],
            [ironbankroller, 'getAllMarkets'],
        )
        names = ['compound', 'cream', 'ironbank']
        return dict(zip(names, results))
    elif chain.id == 56: # bsc mainnet
        venustroller = Contract("0xfD36E2c2a6789Db23113685031d7F16329158384")
        results = [venustroller.getAllMarkets()]
        names = ['venus']
        return dict(zip(names, results))
    elif chain.id == 137: # poly mainnet
        easytroller = Contract("0xcb3fA413B23b12E402Cfcd8FA120f983FB70d8E8")
        try:
            results = [easytroller.getAllMarkets()]
        except:
            from ypricemagic.interfaces.compound.unitroller import UNITROLLER_ABI
            easytroller = Contract.from_abi('Unitroller',"0xcb3fA413B23b12E402Cfcd8FA120f983FB70d8E8", UNITROLLER_ABI)
            results = [easytroller.getAllMarkets()]
        names = ['easyfi']
        return dict(zip(names, results))


@memory.cache()
def is_compound_market(token):
    markets = get_markets()
    if any(token in market for market in markets.values()):
        return True
    # NOTE: Workaround for pools that have since been revoked
    token_contract = Contract(token)
    required = {"isCToken", "comptroller", "underlying"} 
    return set(token_contract.__dict__) & required == required


def get_price(token_address: str, block=None):
    token = Contract(token_address)
    if chain.id in [1,56,137]:
        try:
            underlying, exchange_rate, decimals = fetch_multicall(
                [token, 'underlying'],
                [token, 'exchangeRateCurrent'],
                [token, 'decimals'],
                block=block
            )
            exchange_rate /= 1e18
            under_decimals = _decimals(underlying,block)
            return [exchange_rate * 10 ** (decimals - under_decimals), underlying]
        except AttributeError:
            exchange_rate, decimals = fetch_multicall(
                [token, 'exchangeRateCurrent'],
                [token, 'decimals'],
                block=block
            )
            exchange_rate /= 1e18
            under_decimals = 18
            return [exchange_rate * 10 ** (decimals - under_decimals), "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE"]
    elif chain.id in []: # no current chains without multicall2 support, keeping for future use
        try:
            underlying = token.underlying(block_identifier = block)
            exchange_rate = token.exchangeRateStored(block_identifier = block) / 1e18
            decimals = _decimals(token_address,block)
            under_decimals = _decimals(underlying,block)
            return [exchange_rate * 10 ** (decimals - under_decimals), underlying]
        except AttributeError:
            exchange_rate = token.exchangeRateStored(block_identifier = block) / 1e18
            decimals = _decimals(token_address,block)
            under_decimals = 18
            return [exchange_rate * 10 ** (decimals - under_decimals), "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE"]



import logging

from brownie import web3 as w3
from eth_utils import encode_hex
from eth_utils import function_signature_to_4byte_selector as fourbyte
from requests import Session
from requests.adapters import HTTPAdapter
from web3 import HTTPProvider
from web3.middleware import filter

from ypricemagic.utils.cache import memory

logger = logging.getLogger(__name__)

BATCH_SIZE = 10000
CACHED_CALLS = [
    "name()",
    "symbol()",
    "decimals()",
]
CACHED_CALLS = [encode_hex(fourbyte(data)) for data in CACHED_CALLS]


def should_cache(method, params):
    if method == "eth_call" and params[0]["data"] in CACHED_CALLS:
        return True
    if method == "eth_getCode" and params[1] == "latest":
        return True
    if method == "eth_getLogs":
        return int(params[0]["toBlock"], 16) - int(params[0]["fromBlock"], 16) == BATCH_SIZE - 1
    return False


def cache_middleware(make_request, w3):
    def middleware(method, params):
        logger.debug("%s %s", method, params)

        if should_cache(method, params):
            response = memory.cache(make_request)(method, params)
        else:
            response = make_request(method, params)

        return response

    return middleware


def setup_middleware():
    # patch web3 provider with more connections and higher timeout
    if w3.provider:
        assert w3.provider.endpoint_uri.startswith("http"), "only http and https providers are supported"
        adapter = HTTPAdapter(pool_connections=100, pool_maxsize=100)
        session = Session()
        session.mount("http://", adapter)
        session.mount("https://", adapter)
        w3.provider = HTTPProvider(w3.provider.endpoint_uri, {"timeout": 600}, session)

    # patch and inject local filter middleware
    filter.MAX_BLOCK_REQUEST = BATCH_SIZE
    w3.middleware_onion.add(filter.local_filter_middleware)
    w3.middleware_onion.add(cache_middleware)



from brownie import convert, web3
from eth_utils import encode_hex
from eth_utils import function_signature_to_4byte_selector as fourbyte


def _decimals(contract_address, block=None):
    data = raw_call(contract_address, "decimals()", block=block)
    return convert.to_int(data)


def _symbol(contract_address, block=None):
    data = raw_call(contract_address, "symbol()", block=block)
    return convert.to_string(data) 


def _totalSupply(contract_address, block=None):
    data = raw_call(contract_address, "totalSupply()", block=block)
    return convert.to_int(data)


def _totalSupplyReadable(contract_address, block=None):
    return _totalSupply(contract_address,block) / 10 ** _decimals(contract_address,block)


def raw_call(contract_address: str, method: str, block=None):
    address = convert.to_address(contract_address)
    data = {'to': address,'data': encode_hex(fourbyte(method))}
    return web3.eth.call(data,block_identifier=block)



import logging

from brownie import Contract, chain, web3
from ypricemagic.interfaces.ERC20 import ERC20ABI
from ypricemagic.utils.cache import memory

logger = logging.getLogger(__name__)


def Contract_erc20(address):
    return Contract.from_abi('ERC20',address,ERC20ABI)


def Contract_with_erc20_fallback(address):
    try:
        contract = Contract(address)
    except (AttributeError, ValueError, IndexError):
        contract = Contract_erc20(address)
    return contract


@memory.cache()
def contract_creation_block(address) -> int:
    """
    Determine the block when a contract was created using binary search.
    NOTE Requires access to historical state. Doesn't account for CREATE2 or SELFDESTRUCT.
    """
    logger.info("contract creation block %s", address)
    height = chain.height
    lo, hi = 0, height
    while hi - lo > 1:
        mid = lo + (hi - lo) // 2
        if web3.eth.get_code(address, block_identifier=mid):
            hi = mid
        else:
            lo = mid
    return hi if hi != height else None



from brownie import Contract, multicall

import ypricemagic.magic
from ypricemagic.utils.raw_calls import _decimals, _totalSupplyReadable


def is_gelato_pool(token_address: str) -> bool:
    pool = Contract(token_address)
    return hasattr(pool, 'gelatoBalance0') and hasattr(pool, 'gelatoBalance1')

def get_price(token_address: str, block=None):
    pool = Contract(token_address)
    with multicall:
        token0 = pool.token0(block_identifier = block)
        token1 = pool.token1(block_identifier = block)
        balance0 = pool.gelatoBalance0(block_identifier = block)
        balance1 = pool.gelatoBalance1(block_identifier = block)
    balance0 /= 10 ** _decimals(token0,block)
    balance1 /= 10 ** _decimals(token1,block)
    totalSupply = _totalSupplyReadable(token_address,block)
    totalVal = balance0 * ypricemagic.magic.get_price(token0,block) + balance1 * ypricemagic.magic.get_price(token1,block)
    return totalVal / totalSupply



import logging

from brownie import Contract, chain
from cachetools.func import ttl_cache
from joblib import Parallel, delayed

from ypricemagic.constants import dai, usdc, wbtc, weth
from ypricemagic.interfaces.balancer.WeightedPool import WeightedPoolABI
from ypricemagic.utils.cache import memory
from ypricemagic.utils.events import decode_logs, get_logs_asap
from ypricemagic.utils.multicall2 import fetch_multicall
from ypricemagic.utils.raw_calls import _decimals

from ypricemagic import magic

if chain.id == 1:
    VAULT_V2 = Contract('0xBA12222222228d8Ba445958a75a0704d566BF2C8')
if chain.id == 250:
    VAULT_V2 = Contract('0x20dd72Ed959b6147912C2e529F0a0C651c33c9ce')

@memory.cache()
def is_balancer_pool_v1(address):
    pool = Contract(address)
    required = {"getCurrentTokens", "getBalance", "totalSupply"}
    return set(pool.__dict__) & required == required

@memory.cache()
def is_balancer_pool_v2(address):
    pool = Contract(address)
    required = {'getPoolId','getPausedState','getSwapFeePercentage'}
    return set(pool.__dict__) & required == required

@memory.cache()
def is_balancer_pool(address):
    return is_balancer_pool_v1(address) or is_balancer_pool_v2(address)


@ttl_cache(ttl=600)
def list_pools_v2(block):
    topics = ['0x3c13bc30b8e878c53fd2a36b679409c073afd75950be43d8858768e956fbc20e']
    events = decode_logs(get_logs_asap(VAULT_V2.address, topics, to_block=block))
    return [(event['poolAddress'],event['poolId']) for event in events]


@ttl_cache(ttl=1800)
def deepest_pool_for_token_v2(token, block):
    pools = list_pools_v2(block)
    pools_info = fetch_multicall(*[[VAULT_V2,'getPoolTokens',poolId] for poolAddress, poolId in pools])
    pools = [poolAddress for poolAddress, poolId in pools]
    pools = zip(pools,pools_info)
    deepest_pool = {'pool': None, 'balance': 0}
    for pool, info in pools:
        if str(info) != "((), (), 0)":
            for i in range(len(info[0])):
                _token, balance = info[0][i], info[1][i]
                try: 
                    name = Contract(pool).__dict__['_build']['contractName']
                except:
                    name = 'proceed'
                if _token == token and balance > deepest_pool['balance'] and name not in ['ConvergentCurvePool','MetaStablePool']:
                    deepest_pool = {'pool': pool, 'balance': balance}
    return deepest_pool['pool']


def _magic_get_price(token, block=None):
    try:
        return magic.get_price(token, block)
    except magic.PriceError:
        return 0


def get_pool_price_v1(token, block=None):
    pool = Contract(token)
    tokens, supply = fetch_multicall([pool, "getCurrentTokens"], [pool, "totalSupply"], block=block)
    tokens = list(tokens)
    logging.debug(f'pool tokens: {tokens}')
    if supply == 0:
        return 0
    supply = supply / 1e18
    balances = list(fetch_multicall(*[[pool, "getBalance", token] for token in tokens], block=block))
    for position, balance in enumerate(balances):
        if balance is None:
            balances[position] = 0
    balances = [balance / 10 ** _decimals(token, block) for balance, token in zip(balances, tokens)]
    logging.debug(f'balancer pool balances: {balances}')
    total = sum([balance * _magic_get_price(token, block=block) for balance, token in zip(balances, tokens)])
    return total / supply

def get_pool_price_v2(pool, block=None):
    pool = Contract(pool)
    poolId, vault, totalSupply, decimals = fetch_multicall([pool, "getPoolId"], [pool, "getVault"], [pool, 'totalSupply'], [pool, 'decimals'], block=block)
    vault = Contract(vault)
    totalSupply = totalSupply / 10 ** decimals
    tokens, balances, lastChangedBlock = vault.getPoolTokens(poolId, block_identifier = block)
    token_decimals = fetch_multicall(*[[Contract(token),'decimals'] for token in tokens], block=block)
    balances = (balance / 10 ** decimal for balance, decimal in zip(balances, token_decimals))
    prices = Parallel(4,'threading')(delayed(magic.get_price)(_token, block) for _token in tokens)
    tvl = sum(balance * price for balance, price in zip(balances, prices))
    price = tvl / totalSupply
    return price
    

def get_token_price_v2(token_address, block=None):
    def query_pool_price(pooladdress, poolid):
        poolcontract = Contract(pooladdress)
        pooltokens = VAULT_V2.getPoolTokens(poolid, block_identifier = block)
        try:
            weights = poolcontract.getNormalizedWeights(block_identifier = block)
        except (AttributeError,ValueError):
            weights = [1 for _ in pooltokens[0]]
        pooltokens = list(zip(pooltokens[0],pooltokens[1], weights))
        wethBal, usdcBal, pairedTokenBal = None, None, None
        for pooltoken, balance, weight in pooltokens:
            if pooltoken == usdc:
                usdcBal = balance
                usdcWeight = weight
            if pooltoken == weth:
                wethBal = balance
                wethWeight = weight
            if pooltoken == token_address:
                tokenBal = balance
                tokenWeight = weight
            if len(pooltokens) == 2 and pooltoken != token_address:
                pairedTokenBal = balance
                pairedTokenWeight = weight
                pairedToken = pooltoken

        tokenPrice = None
        if usdcBal:
            usdcValueUSD = (usdcBal / 10 ** 6) * magic.get_price(usdc, block)
            tokenValueUSD = usdcValueUSD / usdcWeight * tokenWeight
            tokenPrice = tokenValueUSD / (tokenBal / 10 ** _decimals(token_address,block))
        elif wethBal:
            wethValueUSD = (wethBal / 10 ** 18) * magic.get_price(weth, block)
            tokenValueUSD = wethValueUSD / wethWeight * tokenWeight
            tokenPrice = tokenValueUSD / (tokenBal / 10 ** _decimals(token_address,block))
        elif pairedTokenBal:
            pairedTokenValueUSD = (pairedTokenBal / 10 ** _decimals(pairedToken,block)) * magic.get_price(pairedToken, block)
            tokenValueUSD = pairedTokenValueUSD / pairedTokenWeight * tokenWeight
            tokenPrice = tokenValueUSD / (tokenBal / 10 ** _decimals(token_address,block))
        return tokenPrice
    
    deepest_pool = deepest_pool_for_token_v2(token_address, block)
    if not deepest_pool:
        return
    try:
        deepest_pool = Contract(deepest_pool)
    except:
        deepest_pool = Contract.from_abi("Balancer Pool V2", deepest_pool, WeightedPoolABI)
    poolId = deepest_pool.getPoolId(block_identifier = block)
    return query_pool_price(deepest_pool.address, poolId)

def get_token_price_v1(token, block=None):
    def get_output(scale=1):
        def check_against(out, scale=1):
            totalOutput = exchange_proxy.viewSplitExactIn(
                token, out, 10 ** _decimals(token) * scale, 32 # NOTE: 32 is max
                , block_identifier = block
            )['totalOutput']
            return out, totalOutput

        try:
            out, totalOutput = check_against(weth)
        except ValueError:
            try:
                out, totalOutput = check_against(dai)
            except ValueError:
                try:
                    out, totalOutput = check_against(usdc)
                except ValueError:
                    try:
                        out, totalOutput = check_against(wbtc)
                    except ValueError:
                        out = None
                        totalOutput = None
        return out, totalOutput
    if chain.id == 1: # chain.id will always be 1, balancer v1 is only on mainnet
        exchange_proxy = Contract('0x3E66B66Fd1d0b02fDa6C811Da9E0547970DB2f21')

    # NOTE: Reverse lookup
    # NOTE: we might not need this when all is said and done
    logging.debug(f'token: {token}')
    logging.debug(f'token after proxy reverse lookup: {token}')
    price = None
    out, totalOutput = get_output()
    if out:
        price = (totalOutput / 10 ** _decimals(out,block)) * magic.get_price(out, block)
        logging.debug('price found via balancer v1')
        return price
    # Can we get an output if we try smaller size?
    scale = 0.5
    out, totalOutput = get_output(scale) 
    if out:
        price = (totalOutput / 10 ** _decimals(out,block)) * magic.get_price(out, block) / scale
        logging.debug('price found via balancer v1')
        return price
    # How about now? 
    scale = 0.1
    out, totalOutput = get_output(scale)
    if out:
        price = (totalOutput / 10 ** _decimals(out,block)) * magic.get_price(out, block) / scale
        logging.debug('price found via balancer v1')
        return price
    else:
        return
        

@ttl_cache(ttl=600)
def get_price(token, block=None):
    if is_balancer_pool_v1(token):
        return get_pool_price_v1(token, block)

    if is_balancer_pool_v2(token):
        return get_pool_price_v2(token, block)

    price = None    
    # NOTE: Only query v2 if block queried > v2 deploy block plus 100000 blocks
    if (
        (chain.id == 1 and (not block or block > 12272146 + 100000))
        or (chain.id == 250 and (not block or block > 16896080))
        ): # lets get some liquidity before we use this as price source
        price = get_token_price_v2(token, block)
    if not price and chain.id == 1:      
        if not block or block > 10730576:   # v1 registry deploy block
            price = get_token_price_v1(token, block)
    return price






from brownie import ZERO_ADDRESS, Contract, chain
from cachetools.func import ttl_cache
from toolz import take
from ypricemagic import magic
from ypricemagic.constants import dai
from ypricemagic.utils.cache import memory
from ypricemagic.utils.multicall2 import fetch_multicall

# curve registry documentation https://curve.readthedocs.io/registry-address-provider.html
try: # if curve registry not deployed on a chain, skip this stuff
    address_provider = Contract('0x0000000022D53366457F9d5E68Ec105046FC4383')
    curve_registry = Contract(address_provider.get_address(0))
    metapool_factory = None if address_provider.get_address(3) == ZERO_ADDRESS else Contract(address_provider.get_address(3))
except:
    pass

# fold underlying tokens into one of the basic tokens

if chain.id == 1:
    BASIC_TOKENS = {
        dai.address,  # dai
        "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",  # weth
        "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE",  # eth
        "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599",  # wbtc
        "0xD71eCFF9342A5Ced620049e616c5035F1dB98620",  # seur
        "0x514910771AF9Ca656af840dff83E8264EcF986CA",  # link
    }
    OVERRIDES = {
        '0x53a901d48795C58f485cBB38df08FA96a24669D5': {
            'name': 'reth',
            'pool': '0xF9440930043eb3997fc70e1339dBb11F341de7A8',
            'coins': [
                '0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE',  # ETH
                '0x9559Aaa82d9649C7A7b220E7c461d2E74c9a3593',  # rETH
            ],
        }
    }
    CRYPTOPOOLS = {
        '0xcA3d75aC011BF5aD07a98d02f18225F9bD9A6BDF': {
            'pool': '0x80466c64868E1ab14a1Ddf27A676C3fcBE638Fe5',
        }, #TriCrypto
        '0xc4AD29ba4B3c580e6D59105FFf484999997675Ff': {
            'pool': '0xD51a44d3FaE010294C616388b506AcdA1bfAAE46',
        }, # 3Crypto
    }
elif chain.id == 137:
    BASIC_TOKENS = {
        dai.address, 
    }
    OVERRIDES = {}
    '''
    OVERRIDES = {
        '0x53a901d48795C58f485cBB38df08FA96a24669D5': {
            'name': 'reth',
            'pool': '0xF9440930043eb3997fc70e1339dBb11F341de7A8',
            'coins': [
                '0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE',  # ETH
                '0x9559Aaa82d9649C7A7b220E7c461d2E74c9a3593',  # rETH
            ],
        }
    }
    '''
    CRYPTOPOOLS = {
        '0x8096ac61db23291252574D49f036f0f9ed8ab390': {
            'pool': '0x751B1e21756bDbc307CBcC5085c042a0e9AaEf36',
        }, # tricrypto
        '0xbece5d20A8a104c54183CC316C8286E3F00ffC71': {
            'pool': '0x92577943c7aC4accb35288aB2CC84D75feC330aF',
        }, # tricrypto2
        '0xdAD97F7713Ae9437fa9249920eC8507e5FbB23d3': {
            'pool': '0x92215849c439E1f8612b6646060B4E3E5ef822cC',
        }, # tricrypto3
        '0x600743B1d8A96438bD46836fD34977a00293f6Aa': {
            'pool': '0xB446BF7b8D6D4276d0c75eC0e3ee8dD7Fe15783A',
        }, # 
    }
elif chain.id == 250:
    BASIC_TOKENS = {
        dai.address, 
    }
    OVERRIDES = {}
    '''
    OVERRIDES = {
        '0x53a901d48795C58f485cBB38df08FA96a24669D5': {
            'name': 'reth',
            'pool': '0xF9440930043eb3997fc70e1339dBb11F341de7A8',
            'coins': [
                '0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE',  # ETH
                '0x9559Aaa82d9649C7A7b220E7c461d2E74c9a3593',  # rETH
            ],
        }
    }
    '''
    CRYPTOPOOLS = {
        '0x58e57cA18B7A47112b877E31929798Cd3D703b0f': {
            'pool': '0x3a1659Ddcf2339Be3aeA159cA010979FB49155FF',
        },
    }




@memory.cache()
def get_pool(token):
    if token in OVERRIDES:
        return OVERRIDES[token]['pool']
    if token in CRYPTOPOOLS:
        return CRYPTOPOOLS[token]['pool']
    try:
        if set(metapool_factory.get_underlying_coins(token)) != {ZERO_ADDRESS}:
            return token
    except:
        try:
            if set(metapool_factory.get_coins(token)) != {ZERO_ADDRESS}:
                return token
        except:
            pass
    return curve_registry.get_pool_from_lp_token(token)


@memory.cache()
def is_curve_lp_token(token):
    return get_pool(token) != ZERO_ADDRESS


@memory.cache()
def get_underlying_coins(token):
    if token in OVERRIDES:
        return OVERRIDES[token]['coins']
    pool = get_pool(token)
    coins = curve_registry.get_underlying_coins(pool)
    if set(coins) == {ZERO_ADDRESS}:
        try:
            coins = metapool_factory.get_underlying_coins(token)
        except ValueError:
            coins = metapool_factory.get_coins(token)
    return [coin for coin in coins if coin != ZERO_ADDRESS]


def cryptopool_lp_price(token, block=None):
    pool = Contract(CRYPTOPOOLS[token]['pool'])
    token = Contract(token)
    result = fetch_multicall(*[[pool, 'coins', i] for i in range(8)])
    tokens = [Contract(token) for token in result if token]
    n = len(tokens)
    try:
        result = iter(
            fetch_multicall(
                [token, 'totalSupply'],
                *[[token, 'decimals'] for token in tokens],
                *[[pool, 'balances', i] for i in range(n)],
                *[[pool, 'price_oracle', i] for i in range(n - 1)],
                block=block
            )
        )
        supply = next(result) / 1e18
        scales = [10 ** decimals for decimals in take(n, result)]
        balances = [balance / scale for balance, scale in zip(take(n, result), scales)]
        # oracles return price with the first coin as a quote currency
        prices = [1] + [price / 1e18 for price in take(n - 1, result)]
        scale = sum(balance * price for balance, price in zip(balances, prices)) / supply
        return [scale, str(tokens[0])]
    # testing
    except TypeError as e:
        if str(e) == "price_oracle requires no arguments":
            result = iter(
                fetch_multicall(
                    [token, 'totalSupply'],
                    *[[token, 'decimals'] for token in tokens],
                    *[[pool, 'balances', i] for i in range(n)],
                    block=block
                )
            )
            supply = next(result) / 1e18
            scales = [10 ** decimals for decimals in take(n, result)]
            balances = [balance / scale for balance, scale in zip(take(n, result), scales)]
            prices = [magic.get_price(token, block) for token in tokens]
            scale = sum(balance * price for balance, price in zip(balances, prices)) / supply
            return [scale, str(tokens[0])]
    

@ttl_cache(ttl=600)
def get_pool_price(token, block=None):
    if token in CRYPTOPOOLS:
        return cryptopool_lp_price(token, block)

    coins = get_underlying_coins(token)
    try:
        coin = (set(coins) & BASIC_TOKENS).pop()
    except KeyError:
        coin = coins[0]

    # there is a registry.get_virtual_price_from_lp_token,
    # but we call pool in case the registry was not deployed at the block
    pool = Contract(get_pool(token))
    try:
        virtual_price = pool.get_virtual_price(block_identifier=block) / 1e18
    except:
        pool = Contract.from_explorer(get_pool(token))
        virtual_price = pool.get_virtual_price(block_identifier=block) / 1e18
    return [virtual_price, coin]



from brownie import network
import os

if not network.is_connected():
    try:
        network_name = os.environ['BROWNIE_NETWORK_ID']
    except KeyError:
        raise KeyError('In order to use pricemagic outside of a brownie project directory, you will need to set $BROWNIE_NETWORK_ID environment variable with the id of your preferred brownie network connection.')
    network.connect(network_name)




from brownie import Contract

from ypricemagic import magic


def is_atoken_v1(address):
    contract = Contract(address)
    required = {"getInterestRedirectionAddress", "getRedirectedBalance", "underlyingAssetAddress"}
    return set(contract.__dict__) & required == required

def is_atoken_v2(address):
    contract = Contract(address)
    required = {"ATOKEN_REVISION", "UNDERLYING_ASSET_ADDRESS"}
    return set(contract.__dict__) & required == required

def is_atoken(address):
    return is_atoken_v1(address) or is_atoken_v2(address)

def get_price_v1(token, block=None):
    underlying = Contract(token).underlyingAssetAddress(block_identifier = block)
    return magic.get_price(underlying, block=block)

def get_price_v2(token, block=None):
    underlying = Contract(token).UNDERLYING_ASSET_ADDRESS(block_identifier = block)
    return magic.get_price(underlying, block=block)

def get_price(token, block=None):
    if is_atoken_v1(token):
        return get_price_v1(token, block)
    if is_atoken_v2(token):
        return get_price_v2(token, block)




from brownie.network.contract import Contract

def is_froyo(address):
    return address == '0x4f85Bbf3B0265DCEd4Ec72ebD0358ccCf190F1B3'

def get_price(froyo_address, block=None):
    pool = Contract('0x83E5f18Da720119fF363cF63417628eB0e9fd523')
    return pool.get_virtual_price(block_identifier = block) / 10 ** 18